# bodhi_utils.py - utility functions for dealing with bodhi
#
# Copyright 2010, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#     Will Woods <wwoods@redhat.com>
#     Martin Krizek <mkrizek@redhat.com>

import fedora.client
import time
import sys
import re
from datetime import datetime
from autoqa.config import SingleConfigParser, getbool, autoqa_conf
import ConfigParser

def _init():
    '''Initialize this module'''
    global user, pswd, bodhi
    try:
        server = autoqa_conf.get('resources', 'bodhi_server')
    except ConfigParser.Error, e:
        server = ''

    try:
        fas_conf = SingleConfigParser()
        fas_conf.read('/etc/autoqa/fas.conf')
        fas = fas_conf.get_section('fas')
        user = fas['username']
        pswd = fas['password']
        if not user or not pswd:
            raise IOError
    except (IOError, ConfigParser.Error), e:
        # if there's no fas.conf or it's incomplete, we just log in anonymously.
        user = ''
        pswd = ''

    ops = {'base_url': server} if server else {}
    bodhi = fedora.client.bodhi.BodhiClient(username=user, password=pswd, **ops)

_init()

def bodhitime(timestamp):
    '''Convert timestamp (seconds since Epoch, assumed to be local time) to a
    Bodhi-approved time string ('%Y-%m-%d %H:%M:%S')'''
    return time.strftime('%Y-%m-%d %H:%M:%S',time.localtime(timestamp))

def parse_bodhitime(bodhitimestr):
    '''Convert a Bodhi-approved time string ('%Y-%m-%d %H:%M:%S') to a
    float-represented number of seconds since Epoch (local time)'''
    return time.mktime(time.strptime(bodhitimestr,'%Y-%m-%d %H:%M:%S'))

def query_update(package):
    '''Get the last Bodhi update matching criteria.

       Args:
       package -- package NVR --or-- package name --or-- update title --or-- update ID

       Returns: Most recent Bodhi update object matching criteria, or None if
       no such available.
    '''
    res = bodhi.query(package=package)
    if res['updates']:
        return res['updates'][0]
    else:
        return None

def bodhi_list(params, limit=100):
    '''Perform a bodhi 'list' method call, with pagination handling. This method
       accepts more search params than the standard 'query' method.

       Args:
       params -- dictionary of params that will be passed to the 'list' method call.
                 E.g.: {'package':'foo-1.1.fc14'}
       limit -- the maximum number of results returned

       Returns: List of update objects matching criteria
    '''
    params['tg_paginate_limit'] = limit
    params['tg_paginate_no'] = 1

    updates = list()
    num_items = 1 # this is a lie to get us in the loop
    while num_items > len(updates):
        r = bodhi.send_request('list', params)
        num_items = r['num_items']
        updates += r['updates']
        params['tg_paginate_no'] += 1
    return updates

def bodhi_already_commented(update, testname, arch):
    '''Check if AutoQA comment is already posted.

       Args:
       update -- Bodhi update object --or-- update title --or-- update ID
       testname -- the name of the test
       arch -- tested architecture

       Returns:
       Tuple containing old result and time when the last comment was posted.
       If no comment is posted yet, tuple will contain two empty strings.

       Throws:
       ValueError -- if no such update can be found
    '''
    # if we received update title or ID, let's convert it to update object first
    if isinstance(update, str):
        u = query_update(update)
        if u:
            update = u
        else:
            raise ValueError("No such update: %s" % update)

    comment_re = r'AutoQA:[\s]+%s[\s]+test[\s]+(\w+)[\s]+on[\s]+%s' % (testname, arch)
    old_result = ''
    comment_time = ''

    for comment in update['comments']:
        if comment['author'] == user:
            m = re.match(comment_re, comment['text'])
            if m == None:
                continue
            old_result = m.group(1)
            comment_time = comment['timestamp']

    return (old_result, comment_time)

def _is_bodhi_testresult_needed(old_result, comment_time, result, time_span):
    '''Check if the comment is meant to be posted.

       Args:
       old_result -- the result of the last test
       comment_time -- the comment time of the last test
       result -- the result of the test
       time_span -- waiting period before posting the same comment

       Returns:
       True if the comment will be posted, False otherwise.
    '''
    # the first comment or a comment with different result, post it
    if not old_result or old_result != result:
        return True

    # If we got here, it means that the comment with the same result has been
    # already posted, we now need to determine whether we can post the
    # comment again or not.
    # If the previous result is *not* 'FAILED', we won't post it in order not to
    # spam developers.
    # If the previous result *is* 'FAILED', we will need to check whether given
    # time span expired, if so, we will post the same comment again to remind
    # a developer about the issue.

    if result != 'FAILED':
        return False

    posted_datetime = datetime.strptime(comment_time, '%Y-%m-%d %H:%M:%S')
    delta = (datetime.utcnow() - posted_datetime)
    # total_seconds() is introduced in python 2.7, until 2.7 is everywhere...
    total_seconds = (delta.microseconds + (delta.seconds + delta.days * 24 * 3600) * 10**6) / 10**6
    minutes = total_seconds/60.0
    if minutes < time_span:
        return False

    return True

def bodhi_post_testresult(update, testname, result, url, arch = 'noarch', karma = 0):
    '''Post comment and karma to bodhi

       Args:
       update -- the *title* of the update comment on
       testname -- the name of the test
       result -- the result of the test
       url -- url of the result of the test
       arch -- tested architecture (default 'noarch')
       karma -- karma points (default 0)

       Returns:
       True if comment was posted successfully or comment wasn't meant to be
       posted (either posting is turned off or comment was already posted),
       False otherwise.
    '''

    # TODO when new bodhi releases, add update identification by UPDATEID support

    err_msg = 'Could not post a comment to bodhi'

    if not update or not testname or not result or url == None:
        sys.stderr.write('Incomplete arguments!\n%s\n' % err_msg)
        return False

    try:
        if not getbool(autoqa_conf.get('notifications', 'send_bodhi_comments')):
            raise ValueError
    except ValueError:
        print 'Sending bodhi comments is turned off. Test result will NOT be sent.'
        # it's either False or not a bool -> it's false, do not send it (but return True since
        # it's intentional, not an error)
        return True

    if not user or not pswd:
        sys.stderr.write('Conf file containing FAS credentials is incomplete!\n%s\n' % err_msg)
        return False

    comment = 'AutoQA: %s test %s on %s. The result can be found at: %s.' \
               % (testname, result, arch, url)
    try:
        (old_result, comment_time) = bodhi_already_commented(update, testname, arch)
        time_span = int(autoqa_conf.get('notifications', 'bodhi_posting_comments_span'))

        if not _is_bodhi_testresult_needed(old_result, comment_time, result, time_span):
            print 'The test result already posted to bodhi.'
            return True

        if not bodhi.comment(update, comment, karma):
            sys.stderr.write('%s\n') % err_msg
            return False

        print 'The test result was sent to bodhi successfully.'
    except Exception, e:
        sys.stderr.write('An error occured: %s\n' % e)
        sys.stderr.write('Could not connect to bodhi!\n%s\n' % err_msg)
        return False

    return True

def _self_test():
    '''
    Simple self test.
    '''
    from datetime import timedelta
    time_span = int(autoqa_conf.get('notifications', 'bodhi_posting_comments_span'))
    try:
        print '1. Test:',
        assert _is_bodhi_testresult_needed('PASSED', datetime.now, 'PASSED', time_span) == False
        print 'Passed'
        print '2. Test:',
        assert _is_bodhi_testresult_needed('FAILED', datetime.now, 'PASSED', time_span) == True
        print 'Passed'
        print '3. Test:',
        assert _is_bodhi_testresult_needed('PASSED', datetime.now, 'FAILED', time_span) == True
        print 'Passed'
        print '4. Test:',
        date = (datetime.utcnow() - timedelta(minutes=time_span)).\
        strftime('%Y-%m-%d %H:%M:%S')
        assert _is_bodhi_testresult_needed('FAILED', date, 'FAILED', time_span) == True
        print 'Passed'
        print '5. Test:',
        date = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
        assert _is_bodhi_testresult_needed('FAILED', date, 'FAILED', time_span) == False
        print 'Passed'
        print '6. Test:',
        assert _is_bodhi_testresult_needed('', '', 'FAILED', time_span) == True
        print 'Passed'
    except AssertionError:
        print 'Failed [!!!]'

if __name__ == '__main__':
    _self_test()

