#!/usr/bin/python

# Copyright 2010, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Vojtech Aschenbrenner <vaschenb@redhat.com>

import operator
import autoqa.koji_utils
import rpmUtils.miscutils
from autoqa.repoinfo import repoinfo
from autoqa.test import AutoQATest
from autoqa.decorators import ExceptionCatcher

class upgradepath(AutoQATest):
    version = 1 # increment if setup() changes

    @ExceptionCatcher()
    def initialize(self, config, **kwargs):
        super(self.__class__, self).initialize(config)
        self.result = 'PASSED'
        # order for evaluation of final result; higher index means preference
        self.result_order = ('PASSED','INFO','FAILED','ABORTED')
        self.envr_results = {} # results for invidual packages
        self.outputs = []
        self.highlights = []

    def compare(self, matching_build, tags, op):
        koji = autoqa.koji_utils.SimpleKojiClientSession()
        result = 'PASSED'

        # convert 'op' to human language for error message clarity
        opstr = {operator.eq: 'equal to',
                 operator.ge: 'greater than or equal to',
                 operator.gt: 'greater than',
                 operator.le: 'less than or equal to',
                 operator.lt: 'less than',
                 operator.ne: 'not equal to'}
        assert op in opstr, 'Used unsupported operator: %s' % op

        for tag in tags:
            latest_build = koji.latest_by_tag(tag, matching_build['name'])
            if latest_build is None:
                msg ='{0:<7}{1}'.format('[ OK ]', tag)
                print msg
                self.outputs.append(msg)

                msg = "\tNo such package: %s" % matching_build['name']
                print msg
                self.outputs.append(msg)
            else:
                res = autoqa.koji_utils.compareEVR(matching_build, latest_build)

                if op(res, 0):
                    ok = True
                    msg ='{0:<7}{1}'.format('[ OK ]', tag)
                    print msg
                    self.outputs.append(msg)
                else:
                    ok = False
                    msg ='{0:<7}{1}'.format('[FAIL]', tag)
                    print msg
                    self.outputs.append(msg)
                msg = "\tLatest package: %s" % latest_build['nvr']
                print msg
                self.outputs.append(msg)
                if not ok:
                    msg = '\tError: Requested package must be %s the latest package' % opstr[op]
                    print msg
                    self.outputs.append(msg)
                    result = 'FAILED'
        return result

    @ExceptionCatcher()
    def run_once(self, envrs, kojitag, **kwargs):
        super(self.__class__, self).run_once()
        update_id = kwargs['name'] or kwargs['id']

        # Get a list of all repos we monitor (currently not -testing)
        # FIXME - perhaps we should only query for 'active' repos
        reponames = [reponame for reponame in repoinfo.repos() if not reponame.endswith('-testing')]
        # Get a list of koji-tags
        repotags = [repoinfo.getrepo(reponame)['tag'] for reponame in reponames]
        repotags.sort()
        assert kojitag in repotags, 'Requested unsupported kojitag: %s' % kojitag
        current_tag = repotags.index(kojitag)

        hi_tags = repotags[(current_tag + 1):] # tags higher than current
        low_eq_tags = repotags[:(current_tag + 1)] # tags lower or equal

        koji = autoqa.koji_utils.SimpleKojiClientSession()

        for envr in envrs:
            msg = '%s\n%s into %s\n%s' % (40*'=', envr, kojitag, 40*'=')
            print msg
            self.outputs.append(msg)

            # our testing package
            (name, version, release, epoch, arch) = rpmUtils.miscutils.splitFilename(envr)
            matching_build = {
                    'name': name,
                    'version' : version,
                    'release' : release + '.' + arch,
                    'epoch' : epoch,
                    'envr' : envr,
                    }

            if kojitag.find('updates') < 0 and repotags[current_tag] != repotags[-1]:
                # not *-updates* and not rawhide
                # FIXME: don't warn about pushing into base repo for both rawhide and branched (but possibly only before change deadline?)
                msg = "Warning: Pushing into the base 'fedora' repository"
                print msg
                self.outputs.append(msg)
                if msg not in self.highlights:
                    # this message would be repeated for every package, highlight only once
                    self.highlights.append(msg)

            latest_build = koji.latest_by_tag(kojitag, matching_build['name'])
            if latest_build is not None:
                if autoqa.koji_utils.compareEVR(matching_build, latest_build) <= 0:
                    msg = "Warning: Same or newer build already exists in the same repo: %s" % latest_build['nvr']
                    print msg
                    self.outputs.append(msg)
                    self.highlights.append(msg)

            # compare with lower or equal tags, so version has to be greater or equal
            result = self.compare(matching_build, low_eq_tags, operator.ge)
            # compare with higher tags, so version has to be lower or equal
            result2 = self.compare(matching_build, hi_tags, operator.le)

            # compute pkg result
            if self.result_order.index(result2) > self.result_order.index(result):
                result = result2
            self.envr_results[envr] = result

            msg = 'RESULT: %s' % result
            print msg
            self.outputs.append(msg)

            # update self.result
            if self.result_order.index(result) > self.result_order.index(self.result):
                self.result = result

        # create summary like "1 PASSED, 2 FAILED, 3 INFO"
        summary = []
        for res in self.result_order:
            if res in self.envr_results.values():
                count = len([k for k in self.envr_results.keys() if self.envr_results[k] == res])
                summary.append('%d %s' % (count, res))
        summary = ', '.join(summary)

        # print summary
        msg = 'SUMMARY: ' + summary
        print msg
        self.outputs.append(msg)
        self.highlights.append(msg)

        self.summary = '%s for %s' % (summary, update_id)

